@@ -20,10 +20,10 @@ package io.undertow.servlet.api;
 
 import java.io.File;
 import java.nio.file.Path;
+import java.nio.file.Paths;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
-import java.util.Collections;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
@@ -35,6 +35,7 @@ import java.util.concurrent.Executor;
 
 import javax.servlet.DispatcherType;
 import javax.servlet.MultipartConfigElement;
+import javax.servlet.ServletContextListener;
 import javax.servlet.descriptor.JspConfigDescriptor;
 
 import io.undertow.security.api.AuthenticationMechanism;
@@ -62,14 +63,28 @@ import io.undertow.util.ImmediateAuthenticationMechanismFactory;
  */
 public class DeploymentInfo implements Cloneable {
 
+    private static final int DEFAULT_MAJOR_VERSION;
+
+    static {
+        // UNDERTOW-1810. It is possible at runtime that the class executing this logic has been bytecode
+        // transformed to use a different variant of the Servlet API than it was compiled against,
+        // i.e. EE 9's Servlet 5 instead of EE 8's Servlet 4. Since 4 and 5 are functionally equivalent
+        // except for the package rename, support such a scenario by setting the default major spec
+        // version that is supported based on the package name of a Servlet API class.
+        Package servletPackage = ServletContextListener.class.getPackage();
+        DEFAULT_MAJOR_VERSION = servletPackage.getName().startsWith("jakarta.") ? 5 : 4;
+    }
+
     private String deploymentName;
     private String displayName;
     private String contextPath;
     private ClassLoader classLoader;
     private ResourceManager resourceManager = ResourceManager.EMPTY_RESOURCE_MANAGER;
     private ClassIntrospecter classIntrospecter = DefaultClassIntrospector.INSTANCE;
-    private int majorVersion = 3;
-    private int minorVersion;
+    private int majorVersion = DEFAULT_MAJOR_VERSION;
+    private int minorVersion = 0;
+    private int containerMajorVersion = DEFAULT_MAJOR_VERSION;
+    private int containerMinorVersion = 0;
     private Executor executor;
     private Executor asyncExecutor;
     private Path tempDir;
@@ -90,6 +105,8 @@ public class DeploymentInfo implements Cloneable {
     private int defaultCookieVersion = 0;
     private SessionPersistenceManager sessionPersistenceManager;
     private String defaultEncoding;
+    private String defaultRequestEncoding;
+    private String defaultResponseEncoding;
     private String urlEncoding = null;
     private boolean ignoreFlush = false;
     private AuthorizationManager authorizationManager = DefaultAuthorizationManager.INSTANCE;
@@ -103,6 +120,7 @@ public class DeploymentInfo implements Cloneable {
     private boolean escapeErrorMessage = true;
     private boolean sendCustomReasonPhraseOnError = false;
     private boolean useCachedAuthenticationMechanism = true;
+    private boolean preservePathOnForward = true;
     private AuthenticationMode authenticationMode = AuthenticationMode.PRO_ACTIVE;
     private ExceptionHandler exceptionHandler;
     private final Map<String, ServletInfo> servlets = new HashMap<>();
@@ -189,6 +207,8 @@ public class DeploymentInfo implements Cloneable {
 
     private boolean checkOtherSessionManagers = true;
 
+    private final List<ServletContextListener> deploymentCompleteListeners = new ArrayList<>();
+
     /**
      * A map of content encoding to file extension for pre compressed resource (e.g. gzip -> .gz)
      */
@@ -357,7 +377,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, ServletInfo> getServlets() {
-        return Collections.unmodifiableMap(servlets);
+        return servlets;
     }
 
 
@@ -381,7 +401,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, FilterInfo> getFilters() {
-        return Collections.unmodifiableMap(filters);
+        return filters;
     }
 
     public DeploymentInfo addFilterUrlMapping(final String filterName, final String mapping, DispatcherType dispatcher) {
@@ -407,7 +427,7 @@ public class DeploymentInfo implements Cloneable {
     public List<FilterMappingInfo> getFilterMappings() {
         final ArrayList<FilterMappingInfo> ret = new ArrayList<>(filterUrlMappings);
         ret.addAll(filterServletNameMappings);
-        return Collections.unmodifiableList(ret);
+        return ret;
     }
 
 
@@ -448,21 +468,36 @@ public class DeploymentInfo implements Cloneable {
         return this;
     }
 
-    public DeploymentInfo addServletContainerInitalizer(final ServletContainerInitializerInfo servletContainerInitializer) {
+    public DeploymentInfo addServletContainerInitializer(final ServletContainerInitializerInfo servletContainerInitializer) {
         servletContainerInitializers.add(servletContainerInitializer);
         return this;
     }
 
-    public DeploymentInfo addServletContainerInitalizers(final ServletContainerInitializerInfo... servletContainerInitializer) {
+    @Deprecated // UNDERTOW-1375 Method name is misspelled
+    public DeploymentInfo addServletContainerInitalizer(final ServletContainerInitializerInfo servletContainerInitializer) {
+        return addServletContainerInitializer(servletContainerInitializer);
+    }
+
+    public DeploymentInfo addServletContainerInitializers(final ServletContainerInitializerInfo... servletContainerInitializer) {
         servletContainerInitializers.addAll(Arrays.asList(servletContainerInitializer));
         return this;
     }
 
-    public DeploymentInfo addServletContainerInitalizers(final List<ServletContainerInitializerInfo> servletContainerInitializer) {
+    @Deprecated // UNDERTOW-1375 Method name is misspelled
+    public DeploymentInfo addServletContainerInitalizers(final ServletContainerInitializerInfo... servletContainerInitializer) {
+        return addServletContainerInitializers(servletContainerInitializer);
+    }
+
+    public DeploymentInfo addServletContainerInitializers(final List<ServletContainerInitializerInfo> servletContainerInitializer) {
         servletContainerInitializers.addAll(servletContainerInitializer);
         return this;
     }
 
+    @Deprecated // UNDERTOW-1375 Method name is misspelled
+    public DeploymentInfo addServletContainerInitalizers(final List<ServletContainerInitializerInfo> servletContainerInitializers) {
+        return addServletContainerInitializers(servletContainerInitializers);
+    }
+
     public List<ServletContainerInitializerInfo> getServletContainerInitializers() {
         return servletContainerInitializers;
     }
@@ -497,7 +532,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, String> getInitParameters() {
-        return Collections.unmodifiableMap(initParameters);
+        return initParameters;
     }
 
     public DeploymentInfo addServletContextAttribute(final String name, final Object value) {
@@ -506,7 +541,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, Object> getServletContextAttributes() {
-        return Collections.unmodifiableMap(servletContextAttributes);
+        return servletContextAttributes;
     }
 
     public DeploymentInfo addWelcomePage(final String welcomePage) {
@@ -525,7 +560,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<String> getWelcomePages() {
-        return Collections.unmodifiableList(welcomePages);
+        return welcomePages;
     }
 
     public DeploymentInfo addErrorPage(final ErrorPage errorPage) {
@@ -544,7 +579,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<ErrorPage> getErrorPages() {
-        return Collections.unmodifiableList(errorPages);
+        return errorPages;
     }
 
     public DeploymentInfo addMimeMapping(final MimeMapping mimeMappings) {
@@ -563,7 +598,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<MimeMapping> getMimeMappings() {
-        return Collections.unmodifiableList(mimeMappings);
+        return mimeMappings;
     }
 
 
@@ -583,7 +618,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<SecurityConstraint> getSecurityConstraints() {
-        return Collections.unmodifiableList(securityConstraints);
+        return securityConstraints;
     }
 
     public Executor getExecutor() {
@@ -633,6 +668,18 @@ public class DeploymentInfo implements Cloneable {
         return tempDir;
     }
 
+    /**
+     * @return Returns the {@link #getTempDir() temp directory path} if it's
+     * not null, else returns the system level temporary directory path
+     * pointed to by the Java system property {@code java.io.tmpdir}
+     */
+    public Path requireTempPath() {
+        if (tempDir != null) {
+            return tempDir;
+        }
+        return Paths.get(SecurityActions.getSystemProperty("java.io.tmpdir"));
+    }
+
     public DeploymentInfo setTempDir(final File tempDir) {
         this.tempDir = tempDir != null ? tempDir.toPath() : null;
         return this;
@@ -731,7 +778,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Set<String> getSecurityRoles() {
-        return Collections.unmodifiableSet(securityRoles);
+        return securityRoles;
     }
 
     /**
@@ -747,7 +794,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<HandlerWrapper> getOuterHandlerChainWrappers() {
-        return Collections.unmodifiableList(outerHandlerChainWrappers);
+        return outerHandlerChainWrappers;
     }
 
     /**
@@ -762,7 +809,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<HandlerWrapper> getInnerHandlerChainWrappers() {
-        return Collections.unmodifiableList(innerHandlerChainWrappers);
+        return innerHandlerChainWrappers;
     }
 
     public DeploymentInfo addInitialHandlerChainWrapper(final HandlerWrapper wrapper) {
@@ -771,7 +818,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<HandlerWrapper> getInitialHandlerChainWrappers() {
-        return Collections.unmodifiableList(initialHandlerChainWrappers);
+        return initialHandlerChainWrappers;
     }
 
 
@@ -808,7 +855,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<HandlerWrapper> getSecurityWrappers() {
-        return Collections.unmodifiableList(securityWrappers);
+        return securityWrappers;
     }
 
     public DeploymentInfo addNotificationReceiver(final NotificationReceiver notificationReceiver) {
@@ -827,7 +874,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<NotificationReceiver> getNotificationReceivers() {
-        return Collections.unmodifiableList(notificationReceivers);
+        return notificationReceivers;
     }
 
     public ConcurrentMap<String, Object> getServletContextAttributeBackingMap() {
@@ -952,7 +999,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, Set<String>> getPrincipalVersusRolesMap() {
-        return Collections.unmodifiableMap(principalVersusRolesMap);
+        return principalVersusRolesMap;
     }
 
     /**
@@ -1028,7 +1075,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, AuthenticationMechanismFactory> getAuthenticationMechanisms() {
-        return Collections.unmodifiableMap(authenticationMechanisms);
+        return authenticationMechanisms;
     }
 
     /**
@@ -1122,7 +1169,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<LifecycleInterceptor> getLifecycleInterceptors() {
-        return Collections.unmodifiableList(lifecycleInterceptors);
+        return lifecycleInterceptors;
     }
 
     /**
@@ -1166,7 +1213,7 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public List<SessionListener> getSessionListeners() {
-        return Collections.unmodifiableList(sessionListeners);
+        return sessionListeners;
     }
 
     public AuthenticationMode getAuthenticationMode() {
@@ -1291,6 +1338,24 @@ public class DeploymentInfo implements Cloneable {
         return this;
     }
 
+    public String getDefaultRequestEncoding() {
+        return defaultRequestEncoding;
+    }
+
+    public DeploymentInfo setDefaultRequestEncoding(String defaultRequestEncoding) {
+        this.defaultRequestEncoding = defaultRequestEncoding;
+        return this;
+    }
+
+    public String getDefaultResponseEncoding() {
+        return defaultResponseEncoding;
+    }
+
+    public DeploymentInfo setDefaultResponseEncoding(String defaultResponseEncoding) {
+        this.defaultResponseEncoding = defaultResponseEncoding;
+        return this;
+    }
+
     /**
      * Adds a pre compressed resource encoding and maps it to a file extension
      *
@@ -1305,7 +1370,50 @@ public class DeploymentInfo implements Cloneable {
     }
 
     public Map<String, String> getPreCompressedResources() {
-        return Collections.unmodifiableMap(preCompressedResources);
+        return preCompressedResources;
+    }
+
+    public int getContainerMajorVersion() {
+        return containerMajorVersion;
+    }
+
+    public DeploymentInfo setContainerMajorVersion(int containerMajorVersion) {
+        this.containerMajorVersion = containerMajorVersion;
+        return this;
+    }
+
+    public int getContainerMinorVersion() {
+        return containerMinorVersion;
+    }
+
+    public DeploymentInfo setContainerMinorVersion(int containerMinorVersion) {
+        this.containerMinorVersion = containerMinorVersion;
+        return this;
+    }
+
+    public boolean isPreservePathOnForward() {
+        return preservePathOnForward;
+    }
+
+    public void setPreservePathOnForward(boolean preservePathOnForward) {
+        this.preservePathOnForward = preservePathOnForward;
+    }
+
+    /**
+     * Add's a listener that is only invoked once all other deployment steps have been completed
+     *
+     * The listeners <code>contextDestroyed</code> method will be called after all undeployment steps are undertaken
+     *
+     * @param servletContextListener
+     * @return
+     */
+    public DeploymentInfo addDeploymentCompleteListener(ServletContextListener servletContextListener) {
+        deploymentCompleteListeners.add(servletContextListener);
+        return this;
+    }
+
+    public List<ServletContextListener> getDeploymentCompleteListeners() {
+        return deploymentCompleteListeners;
     }
 
     @Override
@@ -1369,7 +1477,9 @@ public class DeploymentInfo implements Cloneable {
         info.invalidateSessionOnLogout = invalidateSessionOnLogout;
         info.defaultCookieVersion = defaultCookieVersion;
         info.sessionPersistenceManager = sessionPersistenceManager;
-        info.principalVersusRolesMap.putAll(principalVersusRolesMap);
+        for (Map.Entry<String, Set<String>> e : principalVersusRolesMap.entrySet()) {
+            info.principalVersusRolesMap.put(e.getKey(), new HashSet<>(e.getValue()));
+        }
         info.ignoreFlush = ignoreFlush;
         info.authorizationManager = authorizationManager;
         info.authenticationMechanisms.putAll(authenticationMechanisms);
@@ -1395,7 +1505,13 @@ public class DeploymentInfo implements Cloneable {
         info.securityDisabled = securityDisabled;
         info.useCachedAuthenticationMechanism = useCachedAuthenticationMechanism;
         info.checkOtherSessionManagers = checkOtherSessionManagers;
+        info.defaultRequestEncoding = defaultRequestEncoding;
+        info.defaultResponseEncoding = defaultResponseEncoding;
         info.preCompressedResources.putAll(preCompressedResources);
+        info.containerMajorVersion = containerMajorVersion;
+        info.containerMinorVersion = containerMinorVersion;
+        info.deploymentCompleteListeners.addAll(deploymentCompleteListeners);
+        info.preservePathOnForward = preservePathOnForward;
         return info;
     }
 
